import six
import pickle
from astropy import units as u
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

from twisted.internet.defer import inlineCallbacks

# internal imports
from ..mimic import Mimic
from ..tkutils import addStyle, get_root
from .utils import (plot_compo, INJECTOR_THETA,
                    target_lens_position, PARK_POSITION)
from .. import widgets as w


if not six.PY3:
    import Tkinter as tk
else:
    import tkinter as tk


class COMPOSetupFrame(tk.Frame):
    """
    This is a minimal frame that contains only the buttons for injection side and the pickoff
    angle button.
    """
    def __init__(self, master):

        tk.Frame.__init__(self, master)
        addStyle(self)

        # create control widgets
        tk.Label(self, text='Injection Position').grid(row=0, column=0, pady=4, padx=4, sticky=tk.W)
        self.injection_side = w.Radio(self, ('L', 'R', 'G'), 3, None, initial=1)
        self.injection_side.grid(row=0, column=1, pady=2, stick=tk.W)

        tk.Label(self, text='Pickoff Angle').grid(row=1, column=0, pady=4, padx=4, sticky=tk.W)
        self.pickoff_angle = w.RangedFloat(self, 0.0, -80, 80, None, False,
                                           allowzero=True, width=4)
        self.pickoff_angle.grid(row=1, column=1, pady=2, stick=tk.W)


class COMPOSetupWidget(tk.Toplevel):
    """
    A child window to setup the COMPO pickoff arms.

    This is a minimal frame that contains only the buttons for injection side and the pickoff
    angle button. It is primarily used for defining instrument setups.

    Normally this window is hidden, but can be revealed from the main GUIs menu
    or by clicking on a "use COMPO" widget in the main GUI.
    """
    def __init__(self, parent):
        tk.Toplevel.__init__(self, parent)
        self.transient(parent)
        self.parent = parent

        addStyle(self)
        self.title("COMPO setup")
        # do not display on creation
        # self.withdraw()

        # dont destroy when we click the close button
        self.protocol('WM_DELETE_WINDOW', self.withdraw)

        self.setup_frame = COMPOSetupFrame(self)
        self.setup_frame.pack()

    def dumpJSON(self):
        """
        Encodes current COMPO setup data to JSON compatible dictionary
        """
        raise NotImplementedError

    def loadJSON(self, data):
        """
        Sets widget values from JSON data
        """
        raise NotImplementedError


class COMPOControlWidget(tk.Toplevel):
    """
    A child window to control the COMPO pickoff arms.

    This is a more advanced window that adds widgets to monitor the state of COMPO
    and allow user control of the hardware.

    Normally this window is hidden, but can be revealed from the main GUIs menu
    or by clicking on a "use COMPO" widget in the main GUI.
    """
    def __init__(self, parent):
        tk.Toplevel.__init__(self, parent)
        self.transient(parent)
        self.parent = parent

        g = get_root(self).globals
        addStyle(self)
        self.title("COMPO setup")

        # do not display on creation
        self.withdraw()

        # dont destroy when we click the close button
        self.protocol('WM_DELETE_WINDOW', self.withdraw)

        # frames for sections
        left = tk.Frame(self)
        right = tk.Frame(self)

        self.setup_frame = COMPOSetupFrame(left)
        self.setup_frame.grid(row=1, column=0, columnspan=2, pady=2, sticky=tk.W)

        # buttons
        self.go = w.ActButton(left, width=12, callback=self.move, text='Move')
        self.conn = w.ActButton(left, width=12, callback=self.handle_connection, text='Connect')
        self.stop = w.ActButton(left, width=12, callback=self.stop_all, text='Stop')
        self.home = w.ActButton(left, width=12, callback=self.home_all, text='Home')

        self.conn.grid(row=2, column=0, pady=2, sticky=tk.E)
        self.home.grid(row=2, column=1, pady=2, sticky=tk.W)
        self.stop.grid(row=3, column=1, pady=2, sticky=tk.W)
        self.go.grid(row=3, column=0, pady=2, sticky=tk.E)

        # create status widgets
        status = tk.LabelFrame(left, text='status')
        status.grid(row=4, column=0, columnspan=2, pady=4, padx=4, sticky=tk.N)

        tk.Label(status, text='Injection Arm').grid(row=0, column=0, sticky=tk.W)
        self.injection_status = w.Ilabel(status, text='INIT', width=10, anchor=tk.W)
        self.injection_status.config(bg=g.COL['warn'])
        self.injection_status.grid(row=0, column=1, sticky=tk.W, pady=2, padx=2)

        tk.Label(status, text='Pickoff Arm').grid(row=1, column=0, sticky=tk.W)
        self.pickoff_status = w.Ilabel(status, text='INIT', width=10, anchor=tk.W)
        self.pickoff_status.config(bg=g.COL['warn'])
        self.pickoff_status.grid(row=1, column=1, sticky=tk.W, pady=2, padx=2)

        tk.Label(status, text='Lens Position').grid(row=2, column=0, sticky=tk.W)
        self.lens_status = w.Ilabel(status, text='INIT', width=10, anchor=tk.W)
        self.lens_status.config(bg=g.COL['warn'])
        self.lens_status.grid(row=2, column=1, sticky=tk.W, pady=2, padx=2)

        # telemetry
        tel_frame = tk.LabelFrame(right, text='telemetry')
        self.label = tk.Text(tel_frame, height=10, width=40, bg=g.COL['log'])
        self.label.configure(state=tk.NORMAL, font=g.ENTRY_FONT)
        self.label.pack(fill=tk.Y)
        tel_frame.grid(row=0, column=0, columnspan=1)

        # mimic
        mimic_width = 350
        Mimic.__init__(self, height=int(mimic_width/2.5), width=mimic_width)
        mimic_frame = tk.LabelFrame(right, text='mimic')
        self.canvas = FigureCanvasTkAgg(self.figure, mimic_frame)
        self.canvas.get_tk_widget().pack()
        mimic_frame.grid(row=1, column=0, padx=4, pady=4)

        left.pack(pady=2, side=tk.LEFT, fill=tk.Y)
        right.pack(pady=2, side=tk.LEFT, fill=tk.Y)

    @property
    def session(self):
        return get_root(self).globals.session

    @inlineCallbacks
    def handle_connection(self):
        if not self.session:
            self.print_message('no session')
            return

        if self.conn['text'].lower() == 'disconnect':
            rpc = 'hipercam.compo.rpc.connection.disconnect'
        else:
            rpc = 'hipercam.compo.rpc.connection.connect'
        yield self.session.call(rpc)

    @inlineCallbacks
    def home_stage(self, stage):
        if not self.session:
            self.print_message('no session')
            return
        rpc = "hipercam.compo.rpc.{}.home".format(stage)
        yield self.session.call(rpc)

    @inlineCallbacks
    def home_all(self):
        for stage in ('injection', 'pickoff', 'lens'):
            yield self.home_stage(stage)

    @inlineCallbacks
    def stop_all(self):
        if not self.session:
            self.print_message('no session')
            return
        for stage in ('injection', 'pickoff', 'lens'):
            rpc = "hipercam.compo.rpc.{}.stop".format(stage)
            yield self.session.call(rpc)

    @inlineCallbacks
    def move(self):
        """
        Send commands to COMPO
        """
        if not self.session:
            self.print_message('no session')
            return

        if self.setup_frame.injection_side.value() == 'L':
            ia = INJECTOR_THETA
        elif self.setup_frame.injection_side.value() == 'R':
            ia = -INJECTOR_THETA
        else:
            ia = PARK_POSITION

        lens = target_lens_position(
            self.setup_frame.pickoff_angle.value() * u.deg,
            False  # guiding
        ).to_value(u.mm)
        self.session.publish('hipercam.compo.target_pickoff_angle',
                             self.setup_frame.pickoff_angle.value())
        self.session.publish('hipercam.compo.target_injection_angle',
                             ia.to_value(u.deg))
        self.session.publish('hipercam.compo.target_lens_position', lens)
        yield self.session.call('hipercam.compo.rpc.pickoff.move')
        yield self.session.call('hipercam.compo.rpc.injection.move')
        yield self.session.call('hipercam.compo.rpc.lens.move')

    def send_message(self, topic, msg):
        if self.session:
            self.session.publish(topic, msg)

    def print_message(self, msg):
        self.label.delete(1.0, tk.END)
        self.label.insert(tk.END, msg+'\n')

    def set_stage_status(self, stage, telemetry):
        state = telemetry['state'][stage]
        if stage == 'injection':
            widget = self.injection_status
        elif stage == 'pickoff':
            widget = self.pickoff_status
        elif stage == 'lens':
            widget = self.lens_status
        else:
            raise ValueError('unkown stage ' + stage)

        g = get_root(self).globals
        colours = {'inpos': g.COL['start'], 'moving': g.COL['warn'], 'stopped': g.COL['warn'],
                   'init': g.COL['warn'], 'homing': g.COL['warn']}
        matched_state = set(state).intersection(colours.keys())
        if not matched_state:
            print('unhandled state ' + '/'.join(state))
        elif len(matched_state) > 1:
            print('ambiguous state ' + '/'.join(state))
        else:
            state = matched_state.pop()
            c = colours[state]
            widget.config(text=state.upper(), bg=c)

    def update_mimic(self, telemetry):
        """
        Use incoming telemetry to update mimic
        """
        if not telemetry:
            return

        injection_angle, _ = self.get_stage_position(telemetry, 'injection_angle') * u.deg
        pickoff_angle, _ = self.get_stage_position(telemetry, 'pickoff_angle') * u.deg

        self.ax.clear()
        _ = plot_compo(pickoff_angle, injection_angle, self.ax)
        self.ax.set_xlim(-250, 250)
        self.ax.set_aspect('equal')
        self.ax.set_axis_off()
        self.canvas.draw()

    def get_stage_position(self, telemetry, pos_str):
        try:
            pos = telemetry[pos_str]['current'].value
        except AttributeError:
            pos = telemetry[pos_str]['current']

        try:
            targ = telemetry[pos_str]['target'].value
        except AttributeError:
            targ = telemetry[pos_str]['target']
        return pos, targ

    def on_telemetry(self, package_data):
        telemetry = pickle.loads(package_data)
        state = telemetry['state']

        # check for error status
        g = get_root(self).globals
        if 'error' in state['connection']:
            self.lens_status.config(text='ERROR',
                                    bg=g.COL['critical'])
            self.pickoff_status.config(text='ERROR',
                                       bg=g.COL['critical'])
            self.injection_status.config(text='ERROR',
                                         bg=g.COL['critical'])
            self.conn.config(text='Connect')
        elif 'offline' in state['connection']:
            self.lens_status.config(text='DISCONN',
                                    bg=g.COL['critical'])
            self.pickoff_status.config(text='DISCONN',
                                       bg=g.COL['critical'])
            self.injection_status.config(text='DISCONN',
                                         bg=g.COL['critical'])
            self.conn.config(text='Connect')
        else:
            self.conn.config(text='Disconnect')
            for stage in ('injection', 'pickoff', 'lens'):
                self.set_stage_status(stage, telemetry)

        str = f"{telemetry['timestamp'].iso}:\n"
        for stage, pos_str in zip(
                ('injection', 'pickoff', 'lens'),
                ('injection_angle', 'pickoff_angle', 'lens_position')):
            pos, targ = self.get_stage_position(telemetry, pos_str)

            status = '/'.join(state[stage][4:])
            str += f"{stage}: curr={pos:.2f}, targ={targ:.2f}\n{status}\n\n"

        self.print_message(str)
        self.update_mimic(telemetry)

    def dumpJSON(self):
        """
        Encodes current COMPO setup data to JSON compatible dictionary
        """
        raise NotImplementedError

    def loadJSON(self, data):
        """
        Sets widget values from JSON data
        """
        raise NotImplementedError



# TCS access routines
from __future__ import print_function, unicode_literals, absolute_import, division
import subprocess
import warnings
import re
from astropy import coordinates as coord
from astropy import units as u

from hcam_devices.wamp.utils import call


def getWhtTcs():
    cmd = "ssh whtguest@taurus.ing.iac.es "
    cmd += "setenv LD_LIBRARY_PATH /wht/release/Ubuntu1404-64/lib; "
    cmd += "/wht/release/Ubuntu1404-64/bin/ParameterNoticeBoardLister | grep TCS"
    try:
        results = subprocess.check_output(cmd.split(), timeout=10).decode().split('\n')
    except subprocess.TimeoutExpired:
        raise IOError('Checking TCS info timed out')

    tcs_data = dict()
    pattern = 'TCS\.(\w*) -> (\w*) (.*)'
    for result in results:
        if not result.startswith('TCS.'):
            continue

        match = re.match(pattern, result)
        if match is None:
            warnings.warn('no match for {}'.format(result))
            continue

        name, _, value = match.groups()
        value = value.replace('{', '').replace('}', '')
        # convert anything we can to a float
        try:
            value = float(value)
        except ValueError:
            pass
        tcs_data[name] = value

    coo = coord.SkyCoord(tcs_data['RAHHMMSS'] + ' ' + tcs_data['DECDDMMSS'],
                         unit=(u.hour, u.deg))
    ra = coo.ra.deg
    dec = coo.dec.deg
    pa = tcs_data['skyPa']
    focus = tcs_data['focus']
    return ra, dec, pa, focus


def getGtcTcs():

    try:
        hdr = call('hipercam.gtc.rpc.get_telescope_pars')
    except Exception:
        raise RuntimeError('failed to get tcs from GTC')
    try:
        # pars is a list of strings describing tel info in FITS
        # style, each entry in the list is a different class of
        # thing (weather, telescope, instrument etc).
        ra = float(hdr['RADEG'])
        dec = float(hdr['DECDEG'])
        pa = float(hdr['INSTRPA'])
        focus = float(hdr['M2UZ'])
    except ValueError:
        ra, dec, pa, focus = -999, -999, -999, -999

    return ra, dec, pa, focus



from __future__ import print_function, division, unicode_literals, absolute_import
from six.moves.queue import Queue
"""
This module supplies a single class, the Container. The idea is
that the root widget (the GUI mainwindow) has a single Container
property. Since all sub-widgets can get a reference to the root
window, all sub widgets can access the Container and thus communicate,
hence avoiding the use of globals.
"""


class Container:
    """
    A simple class to hold common attributes for communication between widgets.

    Those in CAPITALS are meant to be immutable; lowercase ones get
    updated.

    The meaning of the globals is as follows (uppercase)::

    COL   : colours
    DAY   : number of seconds in a day
    EFAC  : ratio FWHM/sigma for a gaussian
    EXTINCTION : extinction, mags/airmass
    SKY   : sky brightenesses
    TINS  : Telescop/instrument data
    DEFAULT_FONT : standard font (set in drivers)
    MENU_FONT    : font for menus (set in drivers)
    ENTRY_FONT   : font for data entry fields and mutable fields

    SERVER_POST_PATH : url for posting setups to rack server

    and lowercase (all set = None to start with)::

    astro   : astronomical information
    clog    : command log widget. Used to report actions and results
    count   : count & S/N information widget
    cpars   : dictionary of configuration parameters
    fpslide : focal plane slide widget
    info    : information widget
    ipars   : instrument parameters widget (windows sizes etc)
    observe : widget of observing commands
    rlog    : response log widget. Used to report server responses
    rpars   : run parameter widget
    setup   : setup widget
    ccd_hw  : CCD hardware (temps, flow rates etc) monitoring widget
    start_filter : filter at last press of 'Start'
    logfile : file to log messages
    """

    def __init__(self, telescope_names=None):
        # Ratio of FWHM/sigma for a gaussian
        self.EFAC = 2.3548

        # FIFO Queue for retrieving exceptions from
        # threaded operations
        self.FIFO = Queue()

        # Colours
        self.COL = {
            'main':     '#d0d0ff',  # Colour for the surrounds
            'text':     '#000050',  # Text colour
            'debug':    '#a0a0ff',  # Text background for debug messages
            'warn':     '#f0c050',  # Text background for warnings
            'error':    '#ffa0a0',  # Text background for errors
            'critical': '#ff0000',  # Text background for disasters
            'start':    '#00e000',  # Start / Success button colour when enabled
            'stop':     '#ff5050',  # Stop / Failure button colour when enabled
            'startD':   '#d0e0d0',  # Start / Success button colour when disabled
            'stopD':    '#ffe0e0',  # Stop / Failure button colour when disabled
            'log':      '#e0d4ff',  # Logger windows
            }

        # Telescope / instrument info. Most of this is do with estimating count rates
        self.TINS = {
            'WHT': {
                'latitude':   28.7598742,   # latitude degrees, North positive
                'longitude': -17.8793802,   # longitude degrees, East positive
                'elevation':  2327.,     # Elevation above sea level, metres
                'plateScale': 0.3,     # Arcsecs/unbinned pixel
                'zerop': {
                    'u': 25.67,
                    'g': 27.25,
                    'r': 26.65,
                    'i': 26.79,
                    'z': 26.39
                    }
                },
            'GTC': {
                'latitude':   28.762,   # latitude degrees, North positive
                'longitude': -17.87764,   # longitude degrees, East positive
                'elevation':  2300.,     # Elevation above sea level, metres
                'plateScale': 0.081,     # Arcsecs/unbinned pixel (WHT Collimator)
                'zerop': {
                    'u': 27.47,
                    'g': 29.06,
                    'r': 28.45,
                    'i': 28.60,
                    'z': 28.20
                    }
                },
            'TNT': {
                'latitude':   18.574,   # latitude degrees, North positive
                'longitude':  98.482,   # longitude degrees, East positive
                'elevation':  2449.,     # Elevation above sea level, metres
                'plateScale': 0.456,     # Arcsecs/unbinned pixel (WHT Collimator)
                'zerop': {
                    'u': 22.71,
                    'g': 25.25,
                    'r': 25.01,
                    'i': 24.69,
                    'z': 23.81
                    }
                },
            'NTT_CUBE': {
                'latitude': -29.256,   # latitude degrees, North positive
                'longitude': -70.73,   # longitude degrees, East positive
                'elevation': 2347.,     # Elevation above sea level, metres
                'plateScale': 0.354,     # Arcsecs/unbinned pixel (WHT Collimator)
                'zerop': {
                    'u': 24.62,
                    'g': 26.43,
                    'r': 25.77,
                    'i': 25.63,
                    'z': 24.79
                    }
                },
            }
        if telescope_names is not None:
            self.TINS = {k: self.TINS[k] for k in telescope_names}

        # Sky brightness, mags/sq-arsec
        self.SKY = {
            'd': {'u': 22.4, 'g': 22.2, 'r': 21.4, 'i': 20.7, 'z': 20.3},
            'g': {'u': 21.4, 'g': 21.2, 'r': 20.4, 'i': 20.1, 'z': 19.9},
            'b': {'u': 18.4, 'g': 18.2, 'r': 17.4, 'i': 17.9, 'z': 18.3},
            }

        # Extinction per unit airmass
        self.EXTINCTION = {'u': 0.5, 'g': 0.19, 'r': 0.09, 'i': 0.05, 'z': 0.04}

        # Fonts set by drivers.add_style later
        # Default font, e.g. used for fixed labels
        self.DEFAULT_FONT = None

        # Font for menus
        self.MENU_FONT = None

        # Entry font, used for data entry points and mutable
        # information
        self.ENTRY_FONT = None

        # URL for posting setups
        self.SERVER_POST_PATH = "setup"

        # Command log widget. Used to report actions and results
        self.clog = None

        # Response log widget. Used to report server responses
        self.rlog = None

        # Configuration parameter dictionary: configurable options
        self.cpars = None

        # Instrument parameters widget
        self.ipars = None

        # Run parameters widget
        self.rpars = None

        # Astronomical information widget
        self.astro = None

        # Information widget
        self.info = None

        # Focal plane slide widget
        self.fpslide = None

        # Observing widget
        self.observe = None

        # Count rates, S-to-N etc widget
        self.count = None

        # Instrument setup widget
        self.setup = None

        # ccd hw monitor widget
        self.ccd_hw = None

        # Logging file
        self.logfile = None



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This sub-module is destined for common non-package specific utility
# functions that will ultimately be merged into `astropy.utils`

# As far as possible, utils contains classes of generic use, such as
# PosInt for positive integer input. See e.g. 'hcam' for more instrument
# dependent components.

from __future__ import print_function, unicode_literals, absolute_import, division
import six
if not six.PY3:
    import tkFont as font
else:
    from tkinter import font


def get_root(w):
    """
    Simple method to access root for a widget
    """
    next_level = w
    while next_level.master:
        next_level = next_level.master
    return next_level


def addStyle(w):
    """
    Styles the GUI: global fonts and colours.

    Parameters
    ----------
    w : tkinter.tk
        widget element to style
    """
    # access global container in root widget
    root = get_root(w)
    g = root.globals
    fsize = g.cpars['font_size']
    family = g.cpars['font_family']

    # Default font
    g.DEFAULT_FONT = font.nametofont("TkDefaultFont")
    g.DEFAULT_FONT.configure(size=fsize, weight='bold', family=family)
    w.option_add('*Font', g.DEFAULT_FONT)

    # Menu font
    g.MENU_FONT = font.nametofont("TkMenuFont")
    g.MENU_FONT.configure(family=family)
    w.option_add('*Menu.Font', g.MENU_FONT)

    # Entry font
    g.ENTRY_FONT = font.nametofont("TkTextFont")
    g.ENTRY_FONT.configure(size=fsize, family=family)
    w.option_add('*Entry.Font', g.ENTRY_FONT)

    # position and size
    # root.geometry("320x240+325+200")

    # Default colours. Note there is a difference between
    # specifying 'background' with a capital B or lowercase b
    w.option_add('*background', g.COL['main'])
    w.option_add('*HighlightBackground', g.COL['main'])
    w.config(background=g.COL['main'])




from __future__ import absolute_import, division, print_function

import pickle

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib import patches

import six
import numpy as np

from autobahn.wamp.exception import SerializationError

from hcam_widgets.tkutils import get_root
from hcam_widgets.widgets import IntegerEntry
from hcam_widgets.mimic import Mimic
from hcam_devices.models.slide import BLOCK_POS, UNBLOCK_POS

from twisted.internet.defer import inlineCallbacks

if not six.PY3:
    import Tkinter as tk
else:
    import tkinter as tk


class SlideFrame(tk.LabelFrame, Mimic):
    """
    Self-contained widget to deal with the focal plane slide
    """

    def __init__(self, master, show_mimic=True):
        """
        master  : containing widget
        """
        Mimic.__init__(self, height=200, width=100)
        tk.LabelFrame.__init__(
            self, master, text='Focal plane slide', padx=10, pady=4)

        # Top for table of buttons
        top = tk.Frame(self)

        # Define the buttons
        width = 8
        self.connect = tk.Button(top, fg='black', text='conn', width=width,
                                 command=lambda: self.action('connection.connect'))

        self.home = tk.Button(top, fg='black', text='home', width=width,
                              command=lambda: self.action('stage.home'))

        self.block = tk.Button(top, fg='black', text='block', width=width,
                               command=lambda: self.action('block'))

        self.unblock = tk.Button(top, fg='black', text='unblock', width=width,
                                 command=lambda: self.action('unblock'))

        self.gval = IntegerEntry(top, UNBLOCK_POS, None, True, width=4)

        self.goto = tk.Button(top, fg='black', text='goto', width=width,
                              command=lambda: self.action('goto',
                                                          self.gval.value()))

        self.reset = tk.Button(top, fg='black', text='reset', width=width,
                               command=lambda: self.action('reset'))

        self.stop = tk.Button(top, fg='black', text='stop', width=width,
                              command=lambda: self.action('stage.stop'))

        self.enable = tk.Button(top, fg='black', text='enable', width=width,
                                command=lambda: self.action('enable'))

        self.disable = tk.Button(top, fg='black', text='disable', width=width,
                                 command=lambda: self.action('disable'))

        self.restore = tk.Button(top, fg='black', text='restore', width=width,
                                 command=lambda: self.action('restore'))

        # arrange the permanent ones
        self.connect.grid(row=0, column=0)
        self.home.grid(row=0, column=1)
        self.block.grid(row=0, column=2)
        self.unblock.grid(row=1, column=2)
        self.goto.grid(row=1, column=0)
        self.gval.grid(row=1, column=1)

        # set others according to expertlevel
        self.setExpertLevel()

        g = get_root(self).globals

        # widget for messages
        self.label = tk.Text(top, height=4, width=30, bg=g.COL['log'])
        self.label.configure(state=tk.NORMAL, font=g.ENTRY_FONT)
        self.label.grid(row=4, column=0, columnspan=3)

        # mimic
        self.showing_mimic = show_mimic
        if show_mimic:
            self.canvas = FigureCanvasTkAgg(self.figure, top)
            self.canvas.get_tk_widget().grid(row=5, column=0, columnspan=3)

        top.pack(pady=2)

    def update_mimic(self, telemetry):
        """
        Use current telemetry to update Mimic of the slide
        """
        if not telemetry or not self.showing_mimic:
            return

        try:
            pos = telemetry['position']['current'].value
        except Exception:
            pos = telemetry['position']['current']

        state = telemetry['state']
        connection_state = state['connection']
        stage_state = state['stage']

        ccd_color, slide_color = ('b', 'k')  # chip slide (stopped)
        if ('error' in connection_state or 'offline' in connection_state
                or 'homed' not in stage_state):
            slide_color = 'r'
        elif 'moving' in stage_state:
            slide_color = 'y'

        ll = (0, 0)
        width = 2048
        height = 1024
        ccd = patches.Rectangle(ll, width, height, color=ccd_color, fill=False)
        slide_origin = 3500
        slide_corners = np.array([
            ((-50, slide_origin + 50)),  # top left
            (-50, pos),  # bottom left
            (2048+50, pos),  # bottom right
            (2048+50, slide_origin + 50)  # top right
        ])
        slide = patches.Polygon(slide_corners, closed=False, alpha=0.8, color=slide_color)
        self.ax.clear()
        self.ax.add_patch(ccd)
        self.ax.add_patch(slide)
        self.ax.set_xlim(-80, 2048 + 80)
        self.ax.set_ylim(-200, slide_origin + 100)
        self.ax.set_aspect('equal')
        self.ax.set_axis_off()
        self.canvas.draw()

    def setExpertLevel(self):
        """
        Modifies widget according to expertise level, which in this
        case is just matter of hiding or revealing the LED option
        and changing the lower limit on the exposure button.
        """
        g = get_root(self).globals
        level = g.cpars['expert_level']
        if level == 0:
            self.reset.grid_forget()
            self.enable.grid_forget()
            self.disable.grid_forget()
            self.restore.grid_forget()
            self.stop.grid_forget()
        else:
            self.stop.grid(row=2, column=0)
            self.disable.grid(row=2, column=1)
            self.enable.grid(row=2, column=2)
            self.reset.grid(row=3, column=0)
            self.restore.grid(row=3, column=1)

    @inlineCallbacks
    def action(self, *comm):
        """
        Send a command to the focal plane slide
        """
        session = get_root(self).globals.session
        if not session:
            self.print_message('no session')
            return

        if comm[0] == 'block':
            topic = "hipercam.slide.rpc.stage.move"
            self.set_slide_target_position(BLOCK_POS)
        elif comm[0] == 'unblock':
            topic = "hipercam.slide.rpc.stage.move"
            self.set_slide_target_position(UNBLOCK_POS)
        elif comm[0] == 'goto':
            topic = "hipercam.slide.rpc.stage.move"
            self.set_slide_target_position(int(comm[1]))
        else:
            topic = f"hipercam.slide.rpc.{comm[0]}"

        try:
            yield session.call(topic)
        except SerializationError:
            pass

    def set_slide_target_position(self, pos):
        topic = "hipercam.slide.target_position"
        session = get_root(self).globals.session
        if session:
            session.publish(topic, pos)

    def send_message(self, topic, msg):
        session = get_root(self).globals.session
        if session:
            session.publish(topic, msg)

    def print_message(self, msg):
        self.label.delete(1.0, tk.END)
        self.label.insert(tk.END, msg+'\n')

    def on_telemetry(self, package_data):
        telemetry = pickle.loads(package_data)

        try:
            pos = telemetry['position']['current'].value
            targ = telemetry['position']['target'].value
        except Exception:
            pos = telemetry['position']['current']
            targ = telemetry['position']['target']

        state = telemetry['state']
        if 'error' in state['connection'] or 'offline' in state['connection']:
            self.connect.config(
                command=lambda: self.action('connection.connect'),
                text='conn')
        else:
            self.connect.config(
                command=lambda: self.action('connection.disconnect'),
                text='disconn')

        str = f"{telemetry['timestamp'].iso}:\n"
        status = "/".join(state['stage'][4:])
        str += f"pos: curr={pos:.0f}, targ={targ:.0f}\n{status}\n\n"
        self.update_mimic(telemetry)
        self.print_message(str)



# Utility to query and set temps via Ethernet on Meerstetter
from __future__ import absolute_import, unicode_literals, print_function, division
import six
import pickle

# GUI imports
from hcam_widgets.widgets import RangedInt
from hcam_widgets.tkutils import get_root, addStyle

from twisted.internet.defer import inlineCallbacks

if not six.PY3:
    import Tkinter as tk
else:
    import tkinter as tk


class CCDTempFrame(tk.LabelFrame):
    """
    Self-contained widget to control CCD temps and reset TECS.
    """
    def __init__(self, master):
        tk.LabelFrame.__init__(
            self, master, text='CCD TECs', padx=10, pady=4
        )
        # Top for table of buttons
        top = tk.Frame(self)

        tk.Label(top, text='CCD1').grid(row=0, column=0)
        tk.Label(top, text='CCD2').grid(row=0, column=1)
        tk.Label(top, text='CCD3').grid(row=0, column=2)
        tk.Label(top, text='CCD4').grid(row=0, column=3)
        tk.Label(top, text='CCD5').grid(row=0, column=4)

        # maps ccd number to meerstetter, meerstetter address
        self.mapping = {
            1: (1, 1), 2: (1, 2), 3: (1, 3),
            4: (2, 1), 5: (2, 2)
        }

        self.temp_entry_widgets = {}
        self.setpoint_displays = {}
        self.reset_buttons = {}
        width = 8
        for i in range(1, 6):
            ival = 5
            self.temp_entry_widgets[i] = RangedInt(
                top, ival, -100, 20, None, True, width=width
            )
            self.temp_entry_widgets[i].grid(row=1, column=i-1)
            self.setpoint_displays[i] = tk.Label(top, text='nan', width=width)
            self.setpoint_displays[i].grid(row=2, column=i-1)
            self.reset_buttons[i] = tk.Button(
                top, fg='black', width=width, text='Reset',
                command=lambda ccd=i: self.reset(ccd))
            self.reset_buttons[i].grid(row=3, column=i-1)

        # bind enter to set value routine
        for i in range(1, 6):
            widget = self.temp_entry_widgets[i]
            widget.unbind('<Return>')
            widget.bind('<Return>', lambda event, ccd=i: self.update(ccd))

        top.pack(pady=2)
        addStyle(self)

        self.telemetry_topics = [
            ('hipercam.meerstetter1.telemetry', lambda data: self.on_telemetry(1, data)),
            ('hipercam.meerstetter2.telemetry', lambda data: self.on_telemetry(2, data))
        ]

    def on_telemetry(self, ms, data):
        g = get_root(self).globals
        try:
            telemetry = pickle.loads(data)
        except Exception as err:
            wrn = 'cannot parse telemetry from Meerstetter {}: {}'
            g.clog.warn(wrn.format(ms, str(err)))
        else:
            if ms == 1:
                widget_numbers = (1, 2, 3)
            else:
                widget_numbers = (4, 5)
            for i, num in enumerate(widget_numbers):
                key = 'target_temperature{}'.format(i+1)
                setpoint = telemetry[key].value
                widget = self.setpoint_displays[num]
                widget.configure(text=str(setpoint))

    @inlineCallbacks
    def update(self, ccd):
        g = get_root(self).globals
        if not g.cpars['ccd_temp_monitoring_on']:
            g.clog.warn('Temperature monitoring disabled. Will not update CCD{}'.format(ccd))
            return
        g.clog.info('Updating CCD{}'.format(ccd))
        widget = self.temp_entry_widgets[ccd]
        val = widget.value()
        g.clog.info('desired setpoint ' + str(val))

        try:
            session = get_root(self).globals.session
            topic = 'hipercam.ccd{}.setpoint'.format(ccd)
            g.clog.info('publishing to topic ' + topic)
            yield session.publish(topic, float(val))
        except Exception:
            g.clog.warn('Unable to update setpoint for CCD{}'.format(ccd))

    @inlineCallbacks
    def reset(self, ccd):
        g = get_root(self).globals
        if not g.cpars['ccd_temp_monitoring_on']:
            g.clog.warn('Temperature monitoring disabled. Will not reset CCD{}'.format(ccd))
            return
        g.clog.info('Resetting TEC on CCD{}'.format(ccd))
        ms, address = self.mapping(ccd)
        try:
            rpc = 'hipercam.meerstetter{}.rpc.reset'.format(ms)
            session = get_root(self).globals.session
            yield session.call(rpc, address)
        except Exception:
            g.clog.warn('Unable to reset TEC {}'.format(ccd))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup
import glob
import os

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'astropy', 'requests', 'twisted', 'hcam_devices', 'matplotlib',
    'scipy', 'autobahn'
]

test_requirements = [
    # TODO: put package test requirements here
]

# Treat everything in scripts except README.rst as a script to be installed
scripts = [fname for fname in glob.glob(os.path.join('scripts', '*'))
           if os.path.basename(fname) != 'README.rst']

setup(
    name='hcam_widgets',
    version='1.0.1',
    description="Common Tkinter widgets for HiPerCAM",
    long_description=readme + '\n\n' + history,
    author="Stuart Littlefair",
    author_email='s.littlefair@shef.ac.uk',
    url='https://github.com/HiPERCAM/hcam_widgets',
    download_url='https://github.com/HiPERCAM/hcam_widgets/archive/v1.0.1.tar.gz',
    packages=[
        'hcam_widgets',
        'hcam_widgets.compo',
        'hcam_widgets.hardware'
    ],
    package_dir={'hcam_widgets':
                 'hcam_widgets'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='hcam_widgets',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements
)
